#include "csurf.h"
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <inttypes.h>

#define CSIDH 0
#define CSURF 1

static __inline__ uint64_t rdtsc(void)
{
    uint32_t hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return lo | (uint64_t) hi << 32;
}

fmpz tmp[TMPS_NUM];

int main(int argc, char **argv) {
  int evals = 100, samples = 5, verbose = 0, ell = 101, bsmin=2, bsmax=0, algo = CSIDH;
  
  int opt;
  while ((opt = getopt(argc, argv, "e:s:vl:m:M:p:h")) != -1) {
    switch (opt) {
    case 'e':
      evals = atoi(optarg);
      break;
    case 's':
      samples = atoi(optarg);
      break;
    case 'l':
      ell = atoi(optarg);
      break;
    case 'm':
      bsmin = atoi(optarg);
      break;
    case 'M':
      bsmax = atoi(optarg);
      break;
    case 'p':
      if (optarg[2] == 'U')
	algo = CSURF;
      break;
    case 'v':
      verbose = 1;
      break;
    default:
      fprintf(stderr,
	      "Usage: %s [-v] [-e evals] [-s samples] [-l degree] [-m min steps] [-M max steps] [-p CSIDH|CSURF]\n",
	      argv[0]);
      exit(-1);
    }
  }
  if (bsmax == 0)
    bsmax = ell / 2;
  bsmax -= bsmax % 2;
  int dp = (bsmax - bsmin)/2 + 1;
  
  clock_t t0, t1, t_samples[dp][samples];
  uint64_t c0, c1, c_samples[dp][samples];
  
  init_tmps();
  
  flint_rand_t rng;
  flint_randinit(rng);
  
  fmpz_mod_ctx_t p; curve E;
  point P, ker;
  fmpz_t cof;
  curve_init(E);
  if (algo == CSIDH) {
    CSIDH_ctx(p);
    CSIDH_start(E, p);
  } else {
    CSURF_ctx(p);
    CSURF_start(E, p);
  }    
  point_init(P); point_init(ker);
  fmpz_init(cof);
  fmpz_divexact_ui(cof, p->n, ell);
  
  int tot = dp * samples * evals;
  for (int d = 0; d < dp; d++) {
    for (int s = 0; s < samples; s++) {
      t_samples[d][s] = 0;
      c_samples[d][s] = 0;
      for (int e = 0; e < evals; e++) {
	int it = e + evals*(s + samples*d);
	if (verbose && tot >= 100 && it % (tot / 100) == 0) {
	  printf("%2d%%", 100 * it / tot);
	  fflush(stdout);
	  printf("\r\x1b[K");
	}

	do {
	  fmpz_randm(P->X, rng, p->n);
	  fmpz_randm(P->Z, rng, p->n);
	  point_mul(ker, P, E, cof, p);
	} while (fmpz_is_zero(ker->Z));
	
	t0 = clock();
	c0 = rdtsc();
	VeluResultant_set_bs(E, E, ker, ell, P, p, bsmin + 2*d);
	c1 = rdtsc();
	t1 = clock();
	c_samples[d][s] += c1 - c0;
	t_samples[d][s] += t1 - t0;
      }
    }
  }

  printf("# Using %s prime\n", algo == CSIDH ? "CSIDH" : "CSURF");
  printf("# Samples: %d, evals/sample: %d\n", samples, evals);
  printf("# Isogeny degree: %d\n", ell);
  printf("# Babysteps from: %d to: %d\n", bsmin, bsmax);
  printf("# bs\tcycles\t\ttime (ms)\n");
  for (int d = 0; d < dp; d++) {
    for (int s = 0; s < samples; s++) {
      printf("%d\t", bsmin + 2*d);
      printf("%" PRIu64 "\t", (uint64_t) c_samples[d][s] / evals);
      printf("%.3lf\n", 1000. * t_samples[d][s] / CLOCKS_PER_SEC / evals);
    }
  }
  
  exit(0);
}
