#include "csurf.h"
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <inttypes.h>

static __inline__ uint64_t rdtsc(void)
{
    uint32_t hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return lo | (uint64_t) hi << 32;
}

fmpz tmp[TMPS_NUM];

int main(int argc, char **argv) {
  int samples = 15, verbose = 0;
  
  int opt;
  while ((opt = getopt(argc, argv, "s:vh")) != -1) {
    switch (opt) {
    case 's':
      samples = atoi(optarg);
      break;
    case 'v':
      verbose = 1;
      break;
    default:
      fprintf(stderr,
	      "Usage: %s [-v] [-s samples]\n",
	      argv[0]);
      exit(-1);
    }
  }
  
  clock_t t0, t1, t_samples[CSIDH_LEN-2][samples][2];
  uint64_t c0, c1, c_samples[CSIDH_LEN-2][samples][2];
  
  init_tmps();
  
  flint_rand_t rng;
  flint_randinit(rng);
  
  fmpz_mod_ctx_t p; curve E, E1;
  point P, P1, ker;
  fmpz_t cof;
  CSIDH_ctx(p);
  curve_init(E);
  curve_init(E1);
  CSIDH_start(E, p);
  point_init(P); point_init(P1); point_init(ker);
  fmpz_init(cof);
  
  int tot = (CSIDH_LEN-2)*samples;
  for (int i = 0; i < CSIDH_LEN-2; i++) {
    int ell = CSIDH_fact[i+2];
    fmpz_divexact_ui(cof, p->n, ell);
    for (int s = 0; s < samples; s++) {
      int it = s + samples*i;
      if (verbose && tot >= 100 && it % (tot / 100) == 0) {
	printf("%2d%%", 100 * it / tot);
	fflush(stdout);
	printf("\r\x1b[K");
      }

      do {
	fmpz_randm(P->X, rng, p->n);
	fmpz_randm(P->Z, rng, p->n);
	point_mul(ker, P, E, cof, p);
      } while (fmpz_is_zero(ker->Z));
      point_set(P1, P);
	
      t0 = clock();
      c0 = rdtsc();
      VeluRenes(E1, E, ker, ell, P, p);
      c1 = rdtsc();
      t1 = clock();
      c_samples[i][s][1] = c1 - c0;
      t_samples[i][s][1] = t1 - t0;
      
      t0 = clock();
      c0 = rdtsc();
      VeluResultant(E1, E, ker, ell, P1, p);
      c1 = rdtsc();
      t1 = clock();
      c_samples[i][s][0] = c1 - c0;
      t_samples[i][s][0] = t1 - t0;
    }
  }
  
  printf("# Samples: %d\n", samples);
  for (int i = 0; i < CSIDH_LEN-2; i++) {
    printf("l %d cycles ", CSIDH_fact[i+2]);
    for (int s = 0; s < samples; s++)
      printf("%" PRIu64 " ", (uint64_t) c_samples[i][s][0]);
    printf("baseline ");
    for (int s = 0; s < samples; s++)
      printf("%" PRIu64 " ", (uint64_t) c_samples[i][s][1]);
    printf("\n");
    printf("l %d ms ", CSIDH_fact[i+2]);
    for (int s = 0; s < samples; s++)
      printf("%.3lf ", 1000. * t_samples[i][s][0] / CLOCKS_PER_SEC);
    printf("baseline ");
    for (int s = 0; s < samples; s++)
      printf("%.3lf ", 1000. * t_samples[i][s][1] / CLOCKS_PER_SEC);
    printf("\n");
  }
  
  exit(0);
}
