#include "csurf.h"
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <inttypes.h>

#define KEY_MAX 'm'
#define KEY_ALT 'a'
#define KEY_RND 'r'

#define CSIDH 0
#define CSURF 1

static __inline__ uint64_t rdtsc(void)
{
    uint32_t hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return lo | (uint64_t) hi << 32;
}

fmpz tmp[TMPS_NUM];

int main(int argc, char **argv) {
  int keys = 65, samples = 15, verbose = 0, algo = CSIDH;
  char key_type = KEY_RND;
  int threshold = 150;
  
  int opt;
  while ((opt = getopt(argc, argv, "k:s:d:t:a:vh")) != -1) {
    switch (opt) {
    case 'k':
      keys = atoi(optarg);
      break;
    case 's':
      samples = atoi(optarg);
      break;
    case 't':
      threshold = atoi(optarg);
      break;
    case 'v':
      verbose = 1;
      break;
    case 'a':
      if (optarg[2] == 'U' || optarg[2] == 'u')
	algo = CSURF;
      break;
    case 'd':
      if (optarg[0] == 'm') {
	key_type = KEY_MAX;
	break;
      } else if (optarg[0] == 'a') {
	key_type = KEY_ALT;
	break;
      } else if (optarg[0] == 'r') {
	key_type = KEY_RND;
	break;
      } else
	__attribute__ ((fallthrough));
    default:
      fprintf(stderr,
	      "Usage: %s [-v] [-t threshold] [-k keys] [-s samples] [-d max|alt|rand] [-a CSIDH|CSURF]\n",
	      argv[0]);
      exit(-1);
    }
  }
  
  clock_t t0, t1, t_samples[keys][2][samples];
  uint64_t c0, c1, c_samples[keys][2][samples];
  
  init_tmps();
  
  flint_rand_t rng;
  flint_randinit(rng);
  
  fmpz_mod_ctx_t p; curve E;
  short key_buf[CSIDH_LEN];
  const short *key
    = key_type == KEY_MAX ? (algo == CSIDH ? CSIDH_max_sk : CSURF_max_sk)
    : key_type == KEY_ALT ? (algo == CSIDH ? CSIDH_alt_sk : CSURF_alt_sk)
    : key_buf;
  if (algo == CSIDH)
    CSIDH_ctx(p);
  else
    CSURF_ctx(p);
  curve_init(E);

  int tot = samples * keys;
  for (int e = 0; e < keys; e++) {
    if (algo == CSIDH) {
      CSIDH_start(E, p);
      CSIDH_random_sk(key_buf, rng);  // even if we don't use it
    } else {
      CSURF_start(E, p);
      CSURF_random_sk(key_buf, rng);  // even if we don't use it
    }
    
    for (int s = 0; s < samples; s++) {
      int it = s + samples*e;
      if (verbose && tot >= 100 && it % (tot / 100) == 0) {
	printf("%2d%%", 100 * it / tot);
	fflush(stdout);
	printf("\r\x1b[K");
      }
      
      /* New */
      if (algo == CSIDH) {
	t0 = clock();
	c0 = rdtsc();
	CSIDH_action(E, E, key, p, rng, threshold);
	c1 = rdtsc();
	t1 = clock();
      } else {
	t0 = clock();
	c0 = rdtsc();
	CSURF_action(E, E, key, p, rng, threshold);
	c1 = rdtsc();
	t1 = clock();
      }
      c_samples[e][0][s] = c1 - c0;
      t_samples[e][0][s] = t1 - t0;

      /* Old */
      if (algo == CSIDH) {
	t0 = clock();
	c0 = rdtsc();
	CSIDH_action(E, E, key, p, rng, 600);
	c1 = rdtsc();
	t1 = clock();
      } else {
	t0 = clock();
	c0 = rdtsc();
	CSURF_action(E, E, key, p, rng, 600);
	c1 = rdtsc();
	t1 = clock();
      }
      c_samples[e][1][s] = c1 - c0;
      t_samples[e][1][s] = t1 - t0;
    }
  }

  printf("# Benchmarking %s\n", algo == CSIDH ? "CSIDH" : "CSURF");
  printf("# Keys: %d, samples: %d\n", keys, samples);
  printf("# Threshold: %d\n", threshold);
  for (int e = 0; e < keys; e++) {
    printf("action cycles ");
    for (int s = 0; s < samples; s++)
      printf("%" PRIu64 " ", (uint64_t) c_samples[e][0][s]);
    printf("baseline ");
    for (int s = 0; s < samples; s++)
      printf("%" PRIu64 " ", (uint64_t) c_samples[e][1][s]);
    printf("\n");
  }
  for (int e = 0; e < keys; e++) {
    printf("action ms ");
    for (int s = 0; s < samples; s++)
      printf("%.3lf ", 1000. * t_samples[e][0][s] / CLOCKS_PER_SEC);  
    printf("baseline ");
    for (int s = 0; s < samples; s++)
      printf("%.3lf ", 1000. * t_samples[e][1][s] / CLOCKS_PER_SEC);  
    printf("\n");
  }
  exit(0);
}
