#ifndef CSURF_H
#define CSURF_H

#include <stdlib.h>
#include <assert.h>
#include <fmpz.h>
#include <fmpz_mod.h>
#include <fmpz_mod_poly.h>

/******************* Temp vars *******************/

#define TMPS_NUM 5
extern fmpz tmp[TMPS_NUM];

static inline void init_tmps() {
  for (int i = 0; i < TMPS_NUM; i++)
    fmpz_init(tmp + i);
}
static inline void clear_tmps() {
  for (int i = 0; i < TMPS_NUM; i++)
    fmpz_clear(tmp + i);
}

/******************* GF(p) *******************/

typedef fmpz_mod_ctx_t gfp;

static inline int is_square(const fmpz_t x, const gfp p) {
  fmpz_tdiv_q_2exp(tmp+0, p->n, 1);
  fmpz_mod_pow_fmpz(tmp+1, x, tmp+0, p);
  return fmpz_is_one(tmp+1);
}

/******************* Curves *******************/

// A Montgoery curve y² = x³ + Ax² + x
typedef struct {
  fmpz_t A;
  fmpz_t C;
  fmpz_t A2C4;  // (A+2C)/4
} curve_struct;
typedef curve_struct curve[1];

static inline void curve_init(curve E) {
  fmpz_init(E->A); fmpz_init(E->C); fmpz_init(E->A2C4);
}
static inline void curve_clear(curve E) {
  fmpz_clear(E->A); fmpz_clear(E->C); fmpz_clear(E->A2C4);
}
static inline void curve_set(curve res, const curve E) {
  fmpz_set(res->A, E->A); fmpz_set(res->C, E->C);
  fmpz_set(res->A2C4, E->A2C4);
}
static inline int curve_equal(const curve A, const curve B, const gfp p) {
  fmpz_mod_mul(tmp+0, A->A, B->C, p);
  fmpz_mod_mul(tmp+1, A->C, B->A, p);
  return fmpz_equal(tmp+0, tmp+1);
}
static inline void curve_precomp(curve E, const gfp p) {
  fmpz_mod_add(tmp+0, E->C, E->C, p);
  fmpz_mod_add(tmp+1, E->A, tmp, p);
  fmpz_add_ui(tmp+2, p->n, 1);
  fmpz_tdiv_q_2exp(tmp+3, tmp+2, 2);      // 1/4 mod p, assumes p = 3 mod 4
  fmpz_mod_mul(E->A2C4, tmp+1, tmp+3, p);
}

// A point (X:Z)
typedef struct {
  fmpz_t X;
  fmpz_t Z;
} point_struct;
typedef point_struct point[1];

static inline void point_init(point P) {
  fmpz_init_set_ui(P->X, 1);
  fmpz_init_set_ui(P->Z, 0);
}

static inline void point_clear(point P) {
  fmpz_clear(P->X);
  fmpz_clear(P->Z);
}

static inline void point_set(point res, const point P) {
  fmpz_set(res->X, P->X);
  fmpz_set(res->Z, P->Z);
}

static inline int point_equal(const point P, const point Q, const gfp p) {
  fmpz_mod_mul(tmp+0, P->X, Q->Z, p);
  fmpz_mod_mul(tmp+1, P->Z, Q->X, p);
  return fmpz_equal(tmp+0, tmp+1);
}

static inline void normalize(point P, const gfp p) {
  fmpz_mod_inv(tmp, P->Z, p);
  fmpz_mod_mul(P->X, P->X, tmp, p);
  fmpz_set_ui(P->Z, 1);
}
static inline void normalized(point res, const point P, const gfp p) {
  fmpz_mod_inv(tmp, P->Z, p);
  fmpz_mod_mul(res->X, P->X, tmp, p);
  fmpz_set_ui(res->Z, 1);
}

int is_on_curve(const point P, const curve E, const gfp p);
void point_double(point res, const point P, const curve E, const gfp p);
void point_xadd(point res, const point P, const point Q, const point PmQ, const gfp p);
void point_mul(point res, const point P, const curve E, const fmpz_t n, const gfp p);

/******************* Isogenies *******************/

/*
 * The two isogeny algorithms.
 *
 * Output E1 = E/<ker> and P1 = φ(P).
 */
void VeluRenes(curve E1, const curve E, const point ker, int ell, point P, const gfp p);
void VeluResultant_set_bs(curve E1, const curve E, const point ker, int ell, point P,
			  const gfp p, int bs);
void VeluResultant(curve E1, const curve E, const point ker, int ell, point P, const gfp p);

/******************* CSIDH/CSURF *******************/

/*
 * Parameters
 */
#define CSIDH_LEN 75
static const unsigned short CSIDH_fact[CSIDH_LEN] =
  {4, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79,
   83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167,
   173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263,
   269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359, 367,
   373, 587};
static const unsigned char CSIDH_bs[CSIDH_LEN] =
  {0, 0, 2, 2,  2,  4,  4,  4,  4,  6,  6,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,
   10, 10, 12,  12,  12,  12,  12,  14,  14,  16,  16,  16,  14,  14,  14,  16,  16,
    14,  14,  14,  14,  16,  16,  16,  16,  18,  16,  16,  16,  16,  20,  20,  20,  20,
    22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,
    22,  26};
static const unsigned short CSURF_fact[CSIDH_LEN] =
  {8, 9, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79,
   83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167,
   173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263,
   269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 349, 353, 367, 373, 379,
   383, 389};
static const unsigned char CSURF_bs[CSIDH_LEN] =
  {0, 0, 2, 2,  2,  4,  4,  4,  4,  6,  6,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,
   10, 10, 12,  12,  12,  12,  12,  14,  14,  16,  16,  16,  14,  14,  14,  16,  16,
    14,  14,  14,  14,  16,  16,  16,  16,  18,  16,  16,  16,  16,  20,  20,  20,  20,
    22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,  22,
    22,  22};

static inline void CSIDH_ctx(gfp ctx) {
  fmpz_set_ui(tmp, 1);
  for (int i = 0; i < CSIDH_LEN; i++) {
    fmpz_mul_ui(tmp, tmp, CSIDH_fact[i]);
  }
  fmpz_sub_ui(tmp, tmp, 1);
  fmpz_mod_ctx_init(ctx, tmp);
}
static inline void CSURF_ctx(gfp ctx) {
  fmpz_set_ui(tmp, 1);
  for (int i = 0; i < CSIDH_LEN; i++) {
    fmpz_mul_ui(tmp, tmp, CSURF_fact[i]);
  }
  fmpz_sub_ui(tmp, tmp, 1);
  fmpz_mod_ctx_init(ctx, tmp);
}
static inline void CSIDH_start(curve E, const gfp p) {
  // Initialize as y² = x³ ± x
  fmpz_set_ui(E->A, 0);
  fmpz_set_ui(E->C, 1);
  curve_precomp(E, p);
}
#define CSURF_start CSIDH_start

/*
 * Group action
 */
void action(curve E1, const curve E, const short key[CSIDH_LEN],
	    const gfp p, flint_rand_t rng, int threshold,
	    const unsigned short *fact, const unsigned char *bs_tab);
static inline void CSIDH_action(curve E1, const curve E, const short key[CSIDH_LEN],
				const gfp p, flint_rand_t rng, int threshold) {
  action(E1, E, key, p, rng, threshold, CSIDH_fact, CSIDH_bs);
}
static inline void CSURF_action(curve E1, const curve E, const short key[CSIDH_LEN],
				const gfp p, flint_rand_t rng, int threshold) {
  action(E1, E, key, p, rng, threshold, CSURF_fact, CSURF_bs);
}

/*
 * Keys
 */
static const short CSIDH_max_sk[CSIDH_LEN] =
  {0, 5, 5, 5,
   5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
   5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
   5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5};
static const short CSIDH_alt_sk[CSIDH_LEN] =
  {0, 5, -5, 5,
   -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5,
   5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5,
   -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5};
static const short CSURF_max_sk[CSIDH_LEN] =
  {137, 4, 4, 4,
   5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
   5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
   3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3};
static const short CSURF_alt_sk[CSIDH_LEN] =
  {137, 4, -4, 4,
   -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5,
   5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5, -5, 5,
   -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3, 3, -3};

// Not cryptographically secure :)
#define CENTER(n,r) (n%(2*r)-r)
static inline void CSIDH_random_sk(short key[CSIDH_LEN], flint_rand_t state) {
  key[0] = 0;
  for (int i = 1; i < CSIDH_LEN; i++) {
    mp_limb_t n = n_randlimb(state);
    key[i] = CENTER(n, 5);
  }
}
static inline void CSURF_random_sk(short key[CSIDH_LEN], flint_rand_t state) {
  for (int i = 0; i < CSIDH_LEN; i++) {
    mp_limb_t n = n_randlimb(state);
    key[i] =
      i == 0
      ? n % 275 - 138
      : i < 4
      ? CENTER(n, 4)
      : i < 50
      ? CENTER(n, 5)
      : CENTER(n, 3);
  }
}

#endif
