#include "csurf.h"

static inline void Renes_loop(point EE, point PP, const point P, const point Q, const gfp p) {
  fmpz_mod_sub(tmp+0, P->X, P->Z, p);
  fmpz_mod_add(tmp+1, Q->X, Q->Z, p);
  fmpz_mod_mul(EE->Z, EE->Z, tmp+1, p); // C *= Q.X + Q.Z
  fmpz_mod_mul(tmp+2, tmp+0, tmp+1, p); // P.X * Q.X - P.Z * Q.Z + P.X * Q.Z - P.Z * Q.X
  
  fmpz_mod_add(tmp+0, P->X, P->Z, p);
  fmpz_mod_sub(tmp+1, Q->X, Q->Z, p);
  fmpz_mod_mul(EE->X, EE->X, tmp+1, p); // A *= Q.X - Q.Z
  fmpz_mod_mul(tmp+3, tmp+0, tmp+1, p); // P.X * Q.X - P.Z * Q.Z - P.X * Q.Z + P.Z * Q.X
  
  fmpz_mod_add(tmp+0, tmp+2, tmp+3, p);
  fmpz_mod_mul(PP->X, PP->X, tmp+0, p); // X *= 2 (P.X * Q.X - P.Z * Q.Z)
  fmpz_mod_sub(tmp+0, tmp+2, tmp+3, p);
  fmpz_mod_mul(PP->Z, PP->Z, tmp+0, p); // Z *= 2 (P.Z * Q.X - P.X * Q.Z)
}

static inline void Renes_finish(curve E1, point P, const curve E,
				const point EE, const point PP, int ell, const gfp p) {
  fmpz_mod_mul(tmp, PP->X, PP->X, p);
  fmpz_mod_mul(P->X, tmp, P->X, p);
  fmpz_mod_mul(tmp, PP->Z, PP->Z, p);
  fmpz_mod_mul(P->Z, tmp, P->Z, p);

  // Moody-Shumov
  fmpz_mod_mul(tmp+0, EE->X, EE->X, p);
  fmpz_mod_mul(tmp+1, tmp+0, tmp+0, p);
  fmpz_mod_mul(tmp+2, tmp+1, tmp+1, p);
  fmpz_mod_sub(tmp+0, E->A2C4, E->C, p);
  fmpz_mod_pow_ui(tmp+1, tmp+0, ell, p);
  fmpz_mod_mul(tmp+0, tmp+1, tmp+2, p);

  fmpz_mod_mul(tmp+1, EE->Z, EE->Z, p);
  fmpz_mod_mul(tmp+2, tmp+1, tmp+1, p);
  fmpz_mod_mul(tmp+3, tmp+2, tmp+2, p);
  fmpz_mod_pow_ui(tmp+2, E->A2C4, ell, p);
  fmpz_mod_mul(tmp+1, tmp+2, tmp+3, p);

  fmpz_mod_add(tmp+2, tmp+1, tmp+0, p);
  fmpz_mod_add(E1->A, tmp+2, tmp+2, p);
  fmpz_mod_sub(E1->C, tmp+1, tmp+0, p);
  curve_precomp(E1, p);
}

void VeluRenes(curve E1, const curve E, const point ker, int ell, point P, const gfp p) {
  point Q, R, EE, PP, tmpP;
  
  point_init(Q); point_init(R);
  point_init(EE); point_init(PP);
  point_init(tmpP);

  point_set(Q, ker);
  point_double(R, ker, E, p);

  fmpz_set_ui(PP->X, 1); fmpz_set_ui(PP->Z, 1);
  fmpz_set_ui(EE->X, 1); fmpz_set_ui(EE->Z, 1);
  for (int i = (ell-1)/2; i >= 1; i--) {
    Renes_loop(EE, PP, P, Q, p);
    if (i > 1) {
      point_set(tmpP, R);
      point_xadd(R, R, ker, Q, p);
      point_set(Q, tmpP);
    }
  }
  assert(point_equal(Q, R, p));

  Renes_finish(E1, P, E, EE, PP, ell, p);
  
  point_clear(Q); point_clear(R);
  point_clear(EE); point_clear(PP);
  point_clear(tmpP);
}

static inline int inplace_prod(fmpz * polys, int len, int num, fmpz_poly_t tmp, const fmpz_t p) {
  fmpz_poly_fit_length(tmp, len*num);
  int last_len = len;
  for (int space = len; num > 1; space *= 2) {
    int start;
    for (start = 0; start < (num-3)*space; start += 2*space) {
      _fmpz_mod_poly_mul(tmp->coeffs, polys+start, len, polys+start+space, len, p);
      _fmpz_vec_set(polys+start, tmp->coeffs, 2*len - 1);
    }
    int len2 = num % 2 == 1 ? len : last_len;
    _fmpz_mod_poly_mul(tmp->coeffs, polys+start, len, polys+start+space, len2, p);
    _fmpz_vec_set(polys+start, tmp->coeffs, len + len2 - 1);
    
    last_len = num % 2 == 1 ? last_len : len + len2 - 1;
    len = 2*len - 1;
    num = num/2 + num%2;
  }
  return last_len;
}

void VeluResultant_set_bs(curve E1, const curve E, const point ker, int ell, point P,
			  const gfp p, int bs) {
  bs += bs % 2;             // make baby steps even
  int gs = (ell >> 1) / bs; // giant steps = (ell - 1)/(2 bs)
  assert(gs > 0);

  // Construct subproduct tree for baby steps
  point Q, step, diff, tmpQ;
  point_init(Q); point_init(step); point_init(diff); point_init(tmpQ);
  point_set(Q, ker);
  point_double(step, ker, E, p);
  point_set(diff, ker);

  fmpz_poly_struct **tree = _fmpz_mod_poly_tree_alloc(bs >> 1);
  fmpz *leaves = _fmpz_vec_init(bs >> 1);
  for (int i = (bs >> 1) - 1; i >= 0; i--) {
    normalize(Q, p);
    fmpz_set(leaves+i, Q->X);
    if (i > 0) {
      point_set(tmpQ, Q);
      point_xadd(Q, Q, step, diff, p);
      point_set(diff, tmpQ);
    }
  }
  _fmpz_mod_poly_tree_build(tree, leaves, bs >> 1, p->n);
  
  // Construct the giant steps fractions for E and P
  fmpz_set_si(step->X, bs);         // exploiting step->X, since it's not used yet
  point_mul(Q, ker, E, step->X, p); // Q = bs * ker
  point_double(step, Q, E, p);
  point_set(diff, Q);

  fmpz *giants_P = _fmpz_vec_init(gs*3),
    *giants_E_num = _fmpz_vec_init(gs*3),
    *giants_E_den = _fmpz_vec_init(gs*3);
  for (int i = (gs-1)*3; i >= 0; i -= 3) {
    // h_S(P)
    fmpz_mod_mul(tmp+0, P->X, Q->X, p);
    fmpz_mod_mul(tmp+1, P->Z, Q->Z, p);
    fmpz_mod_mul(tmp+2, tmp+0, tmp+1, p);
    fmpz_mod_mul(tmp+3, tmp+2, E->A, p);
    fmpz_mod_add(giants_P+i+1, tmp+3, tmp+3, p); // 2A P.X P.Z Q.X Q.Z
    
    fmpz_mod_add(tmp+2, tmp+0, tmp+1, p); // P.X Q.X + P.Z Q.Z
    fmpz_mod_sub(tmp+0, tmp+0, tmp+1, p); // P.X Q.X - P.Z Q.Z
    
    fmpz_mod_mul(tmp+1, P->Z, Q->X, p);
    fmpz_mod_mul(tmp+3, P->X, Q->Z, p);
    fmpz_mod_add(tmp+4, tmp+1, tmp+3, p); // P.Z Q.X + P.X Q.Z
    fmpz_mod_sub(tmp+1, tmp+1, tmp+3, p); // P.Z Q.X - P.X Q.Z

    fmpz_mod_mul(tmp+3, tmp+2, E->C, p);  // C(P.X Q.X + P.Z Q.Z)
    fmpz_mod_mul(tmp+2, tmp+3, tmp+4, p); // C(P.X P.Z Q.X² + P.X² Q.X Q.Z + P.X P.Z Q.Z² + P.Z² Q.X Q.Z )
    fmpz_mod_add(tmp+3, tmp+2, giants_P+i+1, p);
    fmpz_mod_add(tmp+2, tmp+3, tmp+3, p);
    fmpz_mod_neg(giants_P+i+1, tmp+2, p);

    fmpz_mod_mul(tmp+2, tmp+0, E->C, p);
    fmpz_mod_mul(giants_P+i+2, tmp+2, tmp+0, p); // C(P.X Q.X - P.Z Q.Z)²
    
    fmpz_mod_mul(tmp+2, tmp+1, E->C, p);
    fmpz_mod_mul(giants_P+i, tmp+2, tmp+1, p);   //  C(P.Z Q.X - P.X Q.Z)²

    // h_S(1), h_S(-1)
    fmpz_mod_add(tmp+0, Q->X, Q->Z, p);
    fmpz_mod_sub(tmp+1, Q->X, Q->Z, p);
    fmpz_mod_mul(tmp+2, tmp+0, tmp+0, p); // (X+Z)²
    fmpz_mod_mul(tmp+3, tmp+1, tmp+1, p); // (X-Z)²
    
    fmpz_mod_mul(giants_E_num+i, tmp+3, E->C, p);
    fmpz_set(giants_E_num+i+2, giants_E_num+i);  // C(X-Z)²
    
    fmpz_mod_mul(giants_E_den+i, tmp+2, E->C, p);
    fmpz_set(giants_E_den+i+2, giants_E_den+i);  // C(X+Z)²

    fmpz_mod_mul(tmp+0, Q->X, Q->Z, p);
    fmpz_mod_mul(tmp+1, tmp+0, E->A, p);
    fmpz_mod_add(tmp+0, tmp+1, tmp+1, p);  // 2AXZ
    fmpz_mod_add(tmp+1, giants_E_den+i, tmp+0, p);
    fmpz_mod_add(tmp+2, tmp+1, tmp+1, p);
    fmpz_mod_neg(giants_E_num+i+1, tmp+2, p); // -2( C(X+Z)² + 2AXZ )

    fmpz_mod_add(tmp+1, giants_E_num+i, tmp+0, p);
    fmpz_mod_add(giants_E_den+i+1, tmp+1, tmp+1, p); // 2( C(X-Z)² + 2AXZ )
    
    if (i >= 3) {
      point_set(tmpQ, Q);
      point_xadd(Q, Q, step, diff, p);
      point_set(diff, tmpQ);
    }
  }

  // Accumulate products in giants
  int g_len;
  fmpz_poly_t tmp_prod;
  fmpz_poly_init(tmp_prod);
  g_len = inplace_prod(giants_P, 3, gs, tmp_prod, p->n);
  assert(g_len == 2*gs + 1);
  g_len = inplace_prod(giants_E_num, 3, gs, tmp_prod, p->n);
  assert(g_len == 2*gs + 1);
  g_len = inplace_prod(giants_E_den, 3, gs, tmp_prod, p->n);
  assert(g_len == 2*gs + 1);
  fmpz_poly_clear(tmp_prod);

  // Compute resultants
  point PP, EE;
  point_init(PP); point_init(EE);
  _fmpz_mod_poly_evaluate_fmpz_vec_fast_precomp(leaves, giants_P, g_len, tree , bs >> 1, p->n);
  for (int i = 1; i < bs >> 1; i++)
    fmpz_mod_mul(leaves, leaves, leaves+i, p);
  fmpz_set(PP->X, leaves);
  _fmpz_poly_reverse(giants_P, giants_P, g_len, g_len);
  _fmpz_mod_poly_evaluate_fmpz_vec_fast_precomp(leaves, giants_P, g_len, tree , bs >> 1, p->n);
  for (int i = 1; i < bs >> 1; i++)
    fmpz_mod_mul(leaves, leaves, leaves+i, p);
  fmpz_set(PP->Z, leaves);
  
  _fmpz_mod_poly_evaluate_fmpz_vec_fast_precomp(leaves, giants_E_num, g_len, tree , bs >> 1, p->n);
  for (int i = 1; i < bs >> 1; i++)
    fmpz_mod_mul(leaves, leaves, leaves+i, p);
  fmpz_set(EE->X, leaves);
  _fmpz_mod_poly_evaluate_fmpz_vec_fast_precomp(leaves, giants_E_den, g_len, tree , bs >> 1, p->n);
  for (int i = 1; i < bs >> 1; i++)
    fmpz_mod_mul(leaves, leaves, leaves+i, p);
  fmpz_set(EE->Z, leaves);
  
  _fmpz_vec_clear(giants_P, gs*3);
  _fmpz_vec_clear(giants_E_num, gs*3); _fmpz_vec_clear(giants_E_den, gs*3);
  _fmpz_vec_clear(leaves, bs >> 1); _fmpz_mod_poly_tree_free(tree, bs >> 1);
  
  // Add missing points
  point_double(Q, ker, E, p);
  point_set(step, Q);
  point_double(diff, Q, E, p);
  for (int i = ell - 2*bs*gs; i > 1 ; i -= 2) {
    Renes_loop(EE, PP, P, Q, p);
    if (i > 3) {
      point_set(tmpQ, diff);
      point_xadd(diff, diff, step, Q, p);
      point_set(Q, tmpQ);
    }
  }
  
  Renes_finish(E1, P, E, EE, PP, ell, p);
  
  point_clear(PP); point_clear(EE);
  point_clear(Q); point_clear(step); point_clear(diff); point_init(tmpQ);
}

void VeluResultant(curve E1, const curve E, const point ker, int ell, point P, const gfp p) {
  VeluResultant_set_bs(E1, E, ker, ell, P, p, ell <= 7 ? 2 : ceil(sqrt(ell)));
}
